#include <string.h>
#include <math.h>
#include "../../plug_in.h"
#include "../../spin_scale.h"
#include "../../utils.h"
#include "bokeh.h"
#include "fft.h"

#ifdef __cplusplus
extern "C" {
#endif

static void GetPolygonPoints(
	double (*points)[2],
	int num_vertices,
	double center_x,
	double center_y,
	double radius
)
{
	double radian;
	int i;

	for(i=0; i<num_vertices; i++)
	{
		radian = (90.0 + 360.0 / num_vertices * i) / 180.0 * G_PI;
		points[i][0] = center_x + (radius * cos(radian) + 0.5);
		points[i][1] = center_y - (radius * sin(radian) + 0.5);
	}
}

void InitializeBokeh(BOKEH_DATA* bokeh, LAYER* target)
{
#define DEFAULT_SIDES 5
#define DEFAULT_BLOOM 2
#define DEFAULT_BLOOM_THRESHOLD 192
	(void)memset(bokeh, 0, sizeof(*bokeh));
	bokeh->temp_pixels = (double*)malloc(sizeof(*bokeh->temp_pixels) * target->width * target->height * 4);
	bokeh->width = target->width / 10;
	bokeh->height = target->height / 10;
	bokeh->stride = bokeh->width * 4;
	bokeh->radius = (bokeh->width > bokeh->height) ?
		bokeh->height * 0.5 : bokeh->width * 0.5;
	bokeh->bloom = DEFAULT_BLOOM;
	bokeh->bloom_threshold = DEFAULT_BLOOM_THRESHOLD;
	bokeh->sides = DEFAULT_SIDES;
}

void ReleaseBokeh(BOKEH_DATA* bokeh)
{
	free(bokeh->shape_pixels);
	free(bokeh->temp_pixels);
}

static void MakeBokehShape(BOKEH_DATA* bokeh)
{
	cairo_t *cairo_p;
	cairo_surface_t *surface_p;
	uint8 *pixels;
	int i;

	pixels = (uint8*)calloc(1, bokeh->width * bokeh->height * 4);
	surface_p = cairo_image_surface_create_for_data(pixels, CAIRO_FORMAT_ARGB32,
		bokeh->width, bokeh->height, bokeh->stride);
	cairo_p = cairo_create(surface_p);

	cairo_set_source_rgb(cairo_p, 1, 1, 1);
	switch(bokeh->shape_type)
	{
	case BOKEH_SHAPE_TYPE_CIRCLE:
		cairo_arc(cairo_p, bokeh->width * 0.5, bokeh->height * 0.5,
			(bokeh->width > bokeh->height) ? bokeh->height * 0.5 : bokeh->width * 0.5, 0, 2 * G_PI);
		cairo_fill(cairo_p);
		break;
	case BOKEH_SHAPE_TYPE_INFLATED_CIRCLE:
		{
			double radius = (bokeh->width > bokeh->height) ? bokeh->height * 0.5 : bokeh->width * 0.5;
			double line_width;

			line_width = radius * 0.45;
			radius = radius - line_width * 0.5;
			cairo_set_line_width(cairo_p, line_width);
			cairo_arc(cairo_p, bokeh->width * 0.5, bokeh->height * 0.5, radius, 0, 2 * G_PI);
			cairo_stroke(cairo_p);
		}
		break;
	case BOKEH_SHAPE_TYPE_TRIANGLE:
		cairo_move_to(cairo_p, bokeh->width * 0.5, 0);
		cairo_line_to(cairo_p, 0, bokeh->height);
		cairo_line_to(cairo_p, bokeh->width, bokeh->height);
		cairo_close_path(cairo_p);
		cairo_fill(cairo_p);
		break;
	case BOKEH_SHAPE_TYPE_SQUARE:
		cairo_move_to(cairo_p, bokeh->width * 0.5, 0);
		cairo_line_to(cairo_p, 0, bokeh->height * 0.5);
		cairo_line_to(cairo_p, bokeh->width * 0.5, bokeh->height);
		cairo_line_to(cairo_p, bokeh->width, bokeh->height * 0.5);
		cairo_close_path(cairo_p);
		cairo_fill(cairo_p);
	case BOKEH_SHAPE_TYPE_HEXAGON:
		{
			double root3 = sqrt(3.0);
			cairo_move_to(cairo_p, bokeh->width * 0.5, 0);
			cairo_line_to(cairo_p, bokeh->width * 0.5 - root3 * bokeh->height * 0.25, bokeh->height * 0.25);
			cairo_line_to(cairo_p, bokeh->width * 0.5 - root3 * bokeh->height * 0.25, (bokeh->height * 0.25) * 3.0);
			cairo_line_to(cairo_p, bokeh->width * 0.5, bokeh->height);
			cairo_line_to(cairo_p, bokeh->width * 0.5 + root3 * bokeh->height * 0.25, (bokeh->height * 0.25) * 3.0);
			cairo_line_to(cairo_p, bokeh->width * 0.5 + root3 * bokeh->height * 0.25, bokeh->height * 0.25);
			cairo_close_path(cairo_p);
			cairo_fill(cairo_p);
		}
		break;
	case BOKEH_SHAPE_TYPE_STAR:
		{
			double points[5][2];
			GetPolygonPoints(points, 5, bokeh->width * 0.5,
				bokeh->height * 0.5, (bokeh->width > bokeh->height) ? bokeh->height * 0.5 : bokeh->width * 0.5);
			cairo_move_to(cairo_p, points[0][0], points[0][1]);
			for(i=1; i<5; i++)
			{
				cairo_line_to(cairo_p, points[(i*2)%5][0], points[(i*2)%5][1]);
			}
			cairo_fill(cairo_p);
		}
		break;
	}

	free(bokeh->shape_pixels);
	bokeh->shape_pixels = (double*)malloc(sizeof(*bokeh->shape_pixels) * bokeh->width * bokeh->height * 4);
	for(i=0; i<bokeh->width * bokeh->height * 4; i++)
	{
		bokeh->shape_pixels[i] = pixels[i] * DIV_PIXEL;
	}

	cairo_destroy(cairo_p);
	cairo_surface_destroy(surface_p);
	free(pixels);
}

static double PixelValue(double* pixels)
{
	return ((pixels[0] * pixels[3] + pixels[1] * pixels[3] + pixels[2] * pixels[3]) / 3);
}

static double ImageMod(double a, double b)
{
	int n = (int)(a/b);
	a -= n*b;
	if(a < 0)
	{
		return a + b;
	}
	return a;
}

int Bokeh(
	BOKEH_DATA* bokeh,
	LAYER* target,
	int (*progress_func)(float rate, void* data),
	void* data
)
{
#define DEFAULT_TILE_SIZE 128
#define DRAW_THRESHOLD (30 / 255.0)
	FFT_DATA fft;
	LAYER *temp_layer = target->window->temp_layer;
	const int width = target->width;
	const int height = target->height;
	int rows = 1,	columns = 1;
	int log2_rows = 0,	log2_columns = 0;
	int radius = (int)bokeh->radius;
	int tile_width = DEFAULT_TILE_SIZE;
	int tile_height = DEFAULT_TILE_SIZE;
	uint8 (*rgb)[4];
	float *mask[2];
	float *green_blue[2];
	float *alpha_red[2];
	int shade_center_x, shade_center_y;
	int w, h;

	double poly_angle;
	double poly_scale;
	double r2;
	double r_angle;
	float total = 0;
	float divide;
	float red, green, blue, alpha;
	int row_flip, column_flip;
	int index;

	int adjusted_width = width + radius * 2;
	int adjusted_height = height + radius * 2;

	int tile_x, tile_y;
	int x, y;
	int i, j, k;

	MakeBokehShape(bokeh);

	tile_width = width + 2*radius;
	tile_height = height + 2*radius;

	(void)memcpy(temp_layer->pixels, target->pixels,
		target->stride * target->height);

	while(rows < tile_height)
	{
		rows *= 2;
		log2_rows++;
	}
	while(columns < tile_width)
	{
		columns *= 2;
		log2_columns++;
	}

	w = columns;
	h = rows;

	tile_width = w;
	tile_height = h;

	InitializeFFT(&fft, MAXIMUM(log2_rows, log2_columns));

	rgb = (uint8 (*)[4])calloc(sizeof(*rgb), w*h);
	mask[0] = (float*)calloc(sizeof(*mask[0]), w*h);
	mask[1] = (float*)calloc(sizeof(*mask[1]), w*h);
	green_blue[0] = (float*)calloc(sizeof(*green_blue[0]), w*h);
	green_blue[1] = (float*)calloc(sizeof(*green_blue[1]), w*h);
	alpha_red[0] = (float*)calloc(sizeof(*alpha_red[0]), w*h);
	alpha_red[1] = (float*)calloc(sizeof(*alpha_red[1]), w*h);

	poly_angle = G_PI / bokeh->sides;
	poly_scale = 1 / cos(poly_angle);
	r2 = bokeh->radius * bokeh->radius;
	r_angle = bokeh->angle * G_PI / 180.0;
	shade_center_x = bokeh->width / 2;
	shade_center_y = bokeh->height / 2;

	i = 0;
	for(y=0; y<h; y++)
	{
		for(x=0; x<w; x++)
		{
			double dx = x - w * 0.5f;
			double dy = y - h * 0.5f;
			double r = dx*dx + dy*dy;
			double f;

			tile_x = (int)(bokeh->width * 0.5 + dx);
			tile_y = (int)(bokeh->height * 0.5 + dy);
			if(tile_x >= 0 && tile_x < bokeh->width && tile_y >= 0 && tile_y < bokeh->height)
			{
				r = sqrt(r);
				if(bokeh->shape_pixels[
					tile_y*bokeh->stride + tile_x*4+3] > DRAW_THRESHOLD)
				{
					if(bokeh->sides > 0)
					{
						double a = atan2(dy, dx) + r_angle;
						a = ImageMod(a, poly_angle*2) - poly_angle;
						f = cos(a) * poly_scale;
					}
					else
					{
						f = 1;
					}
					f = (f * r < bokeh->radius) ? 1 : 0;
				}
				else
				{
					f = 0;
				}
			}
			else
			{
				f = 0;
			}
			total += (float)f;

			mask[0][i] = (float)f;
			mask[1][i] = 0;
			i++;
		}

		if(progress_func((float)(y/2) / h, data) != 0)
		{
			return -1;
		}
	}

	divide = 1.0f / total;
	i = 0;
	for(y=0; y<h; y++)
	{
		for(x=0; x<w; x++)
		{
			mask[0][i] *= divide;
			i++;
		}
	}

	FastFourierTransform2D(&fft, mask[0], mask[1], w, h, 1);

	for(tile_y = -radius; tile_y < height; tile_y += tile_height-2*radius)
	{
		for(tile_x = -radius; tile_x < width; tile_x += tile_width-2*radius)
		{
			int tx = tile_x, ty = tile_y;
			int tw = tile_width, th = tile_height;
			int fx = 0, fy = 0;
			if(tx < 0)
			{
				tw += tx;
				fx -= tx;
				tx = 0;
			}
			if(ty < 0)
			{
				th += ty;
				fy -= ty;
				ty = 0;
			}
			if(tx+tw > width)
			{
				tw = width-tx;
			}
			if(ty+th > height)
			{
				th = height-ty;
			}

			for(y=0; y<th; y++)
			{
				(void)memcpy(rgb[fy*w+fx + y*w],
					&temp_layer->pixels[(y+ty)*target->stride+tx*4], tw*4);
			}

			i = 0;
			for(y=0; y<h; y++)
			{
				int image_y = y + tile_y;
				if(image_y < 0)
				{
					j = fy;
				}
				else if(image_y >= height)
				{
					j = fy + th - 1;
				}
				else
				{
					j = y;
				}
				j *= w;

				for(x=0; x<w; x++)
				{
					int image_x = x + tile_x;
					if(image_x < 0)
					{
						k = fx;
					}
					else if(image_x >= width)
					{
						k = fx + tw - 1;
					}
					else
					{
						k = x;
					}
					k += j;

					alpha_red[0][i] = rgb[k][3];
					red = rgb[k][2];
					green = rgb[k][1];
					blue = rgb[k][0];

					if(red > bokeh->bloom_threshold)
					{
						red *= bokeh->bloom;
					}
					if(green > bokeh->bloom_threshold)
					{
						green *= bokeh->bloom;
					}
					if(blue > bokeh->bloom_threshold)
					{
						blue *= bokeh->bloom;
					}

					alpha_red[1][i] = red;
					green_blue[0][i] = green;
					green_blue[1][i] = blue;
					i++;
					k++;
				}
			}

			FastFourierTransform2D(&fft, alpha_red[0], alpha_red[1],
				columns, rows, 1);
			FastFourierTransform2D(&fft, green_blue[0], green_blue[1],
				columns, rows, 1);

			i = 0;
			for(y=0; y<h; y++)
			{
				for(x=0; x<w; x++)
				{
					float real = alpha_red[0][i];
					float image = alpha_red[1][i];
					float real_mask = mask[0][i];
					float image_mask = mask[1][i];
					alpha_red[0][i] = real*real_mask - image*image_mask;
					alpha_red[1][i] = real*image_mask + image*real_mask;

					real = green_blue[0][i];
					image = green_blue[1][i];
					green_blue[0][i] = real*real_mask - image*image_mask;
					green_blue[1][i] = real*image_mask + image*real_mask;
					i++;
				}
			}

			FastFourierTransform2D(&fft, alpha_red[0], alpha_red[1],
				columns, rows, -1);
			FastFourierTransform2D(&fft, green_blue[0], green_blue[1],
				columns, rows, -1);

			row_flip = w >> 1;
			column_flip = h >> 1;
			index = 0;

			for(y=0; y<w; y++)
			{
				int ym = y ^ row_flip;
				int yi = ym * columns;
				for(x=0; x<w; x++)
				{
					int xm = yi + (x ^ column_flip);
					alpha = alpha_red[0][xm];
					red = alpha_red[1][xm];
					green = green_blue[0][xm];
					blue = green_blue[1][xm];

					if(red > 255)
					{
						red = 255;
					}
					if(green > 255)
					{
						green = 255;
					}
					if(blue > 255)
					{
						blue = 255;
					}

					rgb[index][0] = (uint8)blue;
					rgb[index][1] = (uint8)green;
					rgb[index][2] = (uint8)red;
					rgb[index][3] = (uint8)alpha;
					index++;
				}
			}

			tx = tile_x + radius;
			ty = tile_y + radius;
			tw = tile_width - 2*radius;
			th = tile_height - 2*radius;
			if(tx+tw > width)
			{
				tw = width-tx;
			}
			if(ty+th > height)
			{
				th = height-ty;
			}
			for(y=0; y<th; y++)
			{
				(void)memcpy(&target->pixels[(ty+y)*target->stride+tx*4],
					rgb[radius*w+radius + y*w], tw * 4);
			}
		}

		if(progress_func((float)(tile_y/2 + radius) / (height + radius) + 0.5f, data) != 0)
		{
			return -1;
		}
	}

	free(alpha_red[1]);
	free(alpha_red[0]);
	free(green_blue[1]);
	free(green_blue[0]);
	free(mask[1]);
	free(mask[0]);
	free(rgb);

	ReleaseFFT(&fft);

	return 0;
}

static void OnClickedPatternButton(GtkWidget* button, BOKEH_DATA* data)
{
	data->shape_type = (eBOKEH_SHAPE_TYPE)g_object_get_data(G_OBJECT(button), "shape_type");
}

static void CalcBokehRadius(BOKEH_DATA* bokeh)
{
	bokeh->radius = (bokeh->width > bokeh->height) ?
		bokeh->height * 0.5 : bokeh->width * 0.5;
	bokeh->stride = bokeh->width * 4;
}

GtkWidget* BokehFilterSettingWidgetNew(BOKEH_DATA* bokeh, LAYER* target)
{
#define IMAGE_SIZE 48
	APPLICATION *app = target->window->app;
	GtkWidget *vbox = gtk_vbox_new(FALSE, 0);
	GtkWidget *image;
	GtkWidget *layout;
	GtkWidget *frame;
	GtkWidget *control;
	GList *list = NULL;
	GtkAdjustment *adjustment;
	GdkPixbuf *pixbuf;
	cairo_t *cairo_p;
	cairo_surface_t *surface_p;
	uint8 *pixels;
	int i;

	frame = gtk_frame_new("Pattern Size");
	gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
	layout = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), layout);

	adjustment = GTK_ADJUSTMENT(gtk_adjustment_new(bokeh->width, 1,
		target->width, 1, 1, 0));
	control = SpinScaleNew(adjustment, app->labels->make_new.width, 0);
	(void)g_signal_connect(G_OBJECT(adjustment), "value_changed",
		G_CALLBACK(AdjustmentChangeValueCallBackInt), &bokeh->width);
	g_object_set_data(G_OBJECT(adjustment), "changed_callback", (gpointer)CalcBokehRadius);
	g_object_set_data(G_OBJECT(adjustment), "callback_data", (gpointer)bokeh);
	gtk_box_pack_start(GTK_BOX(layout), control, FALSE, FALSE, 0);

	adjustment = GTK_ADJUSTMENT(gtk_adjustment_new(bokeh->height, 1,
		target->height, 1, 1, 0));
	control = SpinScaleNew(adjustment, app->labels->make_new.height, 0);
	(void)g_signal_connect(G_OBJECT(adjustment), "value_changed",
		G_CALLBACK(AdjustmentChangeValueCallBackInt), &bokeh->height);
	g_object_set_data(G_OBJECT(adjustment), "changed_callback", (gpointer)CalcBokehRadius);
	g_object_set_data(G_OBJECT(adjustment), "callback_data", (gpointer)bokeh);
	gtk_box_pack_start(GTK_BOX(layout), control, FALSE, FALSE, 0);

	frame = gtk_frame_new("Shape");
	gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
	layout = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), layout);

	// ~
	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, IMAGE_SIZE, IMAGE_SIZE);
	pixels = gdk_pixbuf_get_pixels(pixbuf);
	surface_p = cairo_image_surface_create_for_data(pixels, CAIRO_FORMAT_ARGB32,
		IMAGE_SIZE, IMAGE_SIZE, gdk_pixbuf_get_rowstride(pixbuf));
	cairo_p = cairo_create(surface_p);
	cairo_set_source_rgb(cairo_p, 1, 1, 1);
	cairo_rectangle(cairo_p, 0, 0, IMAGE_SIZE, IMAGE_SIZE);
	cairo_fill(cairo_p);
	cairo_set_source_rgb(cairo_p, 0, 0, 0);
	cairo_arc(cairo_p, IMAGE_SIZE * 0.5, IMAGE_SIZE * 0.5, IMAGE_SIZE * 0.5, 0, 2 * G_PI);
	cairo_fill(cairo_p);
	cairo_destroy(cairo_p);
	cairo_surface_destroy(surface_p);
	image = gtk_image_new_from_pixbuf(pixbuf);
	control = ToggleRadioButtonNew(&list,
		(void (*)(GtkWidget*, void*))OnClickedPatternButton, (void*)bokeh);
	g_object_set_data(G_OBJECT(control), "shape_type", GINT_TO_POINTER(BOKEH_SHAPE_TYPE_CIRCLE));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(control), TRUE);
	gtk_container_add(GTK_CONTAINER(control), image);
	gtk_box_pack_start(GTK_BOX(layout), control, FALSE, FALSE, 0);

	// ̉~
	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, IMAGE_SIZE, IMAGE_SIZE);
	pixels = gdk_pixbuf_get_pixels(pixbuf);
	surface_p = cairo_image_surface_create_for_data(pixels, CAIRO_FORMAT_ARGB32,
		IMAGE_SIZE, IMAGE_SIZE, gdk_pixbuf_get_rowstride(pixbuf));
	cairo_p = cairo_create(surface_p);
	cairo_set_source_rgb(cairo_p, 1, 1, 1);
	cairo_rectangle(cairo_p, 0, 0, IMAGE_SIZE, IMAGE_SIZE);
	cairo_fill(cairo_p);
	cairo_set_source_rgb(cairo_p, 0, 0, 0);
	{
		double radius = IMAGE_SIZE * 0.5;
		double line_width;

		line_width = radius * 0.45;
		radius = radius - line_width * 0.5;
		cairo_set_line_width(cairo_p, line_width);
		cairo_arc(cairo_p, IMAGE_SIZE * 0.5, IMAGE_SIZE * 0.5, radius, 0, 2 * G_PI);
		cairo_stroke(cairo_p);
	}
	cairo_destroy(cairo_p);
	cairo_surface_destroy(surface_p);
	image = gtk_image_new_from_pixbuf(pixbuf);
	control = ToggleRadioButtonNew(&list,
		(void (*)(GtkWidget*, void*))OnClickedPatternButton, (void*)bokeh);
	g_object_set_data(G_OBJECT(control), "shape_type", GINT_TO_POINTER(BOKEH_SHAPE_TYPE_INFLATED_CIRCLE));
	gtk_container_add(GTK_CONTAINER(control), image);
	gtk_box_pack_start(GTK_BOX(layout), control, FALSE, FALSE, 0);

	// Op`
	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, IMAGE_SIZE, IMAGE_SIZE);
	pixels = gdk_pixbuf_get_pixels(pixbuf);
	surface_p = cairo_image_surface_create_for_data(pixels, CAIRO_FORMAT_ARGB32,
		IMAGE_SIZE, IMAGE_SIZE, gdk_pixbuf_get_rowstride(pixbuf));
	cairo_p = cairo_create(surface_p);
	cairo_set_source_rgb(cairo_p, 1, 1, 1);
	cairo_rectangle(cairo_p, 0, 0, IMAGE_SIZE, IMAGE_SIZE);
	cairo_fill(cairo_p);
	cairo_set_source_rgb(cairo_p, 0, 0, 0);
	cairo_move_to(cairo_p, IMAGE_SIZE * 0.5, 0);
	cairo_line_to(cairo_p, 0, IMAGE_SIZE);
	cairo_line_to(cairo_p, IMAGE_SIZE, IMAGE_SIZE);
	cairo_close_path(cairo_p);
	cairo_fill(cairo_p);
	cairo_destroy(cairo_p);
	cairo_surface_destroy(surface_p);
	image = gtk_image_new_from_pixbuf(pixbuf);
	control = ToggleRadioButtonNew(&list,
		(void (*)(GtkWidget*, void*))OnClickedPatternButton, (void*)bokeh);
	g_object_set_data(G_OBJECT(control), "shape_type", GINT_TO_POINTER(BOKEH_SHAPE_TYPE_TRIANGLE));
	gtk_container_add(GTK_CONTAINER(control), image);
	gtk_box_pack_start(GTK_BOX(layout), control, FALSE, FALSE, 0);

	// H^
	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, IMAGE_SIZE, IMAGE_SIZE);
	pixels = gdk_pixbuf_get_pixels(pixbuf);
	surface_p = cairo_image_surface_create_for_data(pixels, CAIRO_FORMAT_ARGB32,
		IMAGE_SIZE, IMAGE_SIZE, gdk_pixbuf_get_rowstride(pixbuf));
	cairo_p = cairo_create(surface_p);
	cairo_set_source_rgb(cairo_p, 1, 1, 1);
	cairo_rectangle(cairo_p, 0, 0, IMAGE_SIZE, IMAGE_SIZE);
	cairo_fill(cairo_p);
	cairo_set_source_rgb(cairo_p, 0, 0, 0);
	cairo_move_to(cairo_p, IMAGE_SIZE * 0.5, 0);
	cairo_line_to(cairo_p, 0, IMAGE_SIZE * 0.5);
	cairo_line_to(cairo_p, IMAGE_SIZE * 0.5, IMAGE_SIZE);
	cairo_line_to(cairo_p, IMAGE_SIZE, IMAGE_SIZE * 0.5);
	cairo_close_path(cairo_p);
	cairo_fill(cairo_p);
	cairo_destroy(cairo_p);
	cairo_surface_destroy(surface_p);
	image = gtk_image_new_from_pixbuf(pixbuf);
	control = ToggleRadioButtonNew(&list,
		(void (*)(GtkWidget*, void*))OnClickedPatternButton, (void*)bokeh);
	g_object_set_data(G_OBJECT(control), "shape_type", GINT_TO_POINTER(BOKEH_SHAPE_TYPE_SQUARE));
	gtk_container_add(GTK_CONTAINER(control), image);
	gtk_box_pack_start(GTK_BOX(layout), control, FALSE, FALSE, 0);

	// Zp`
	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, IMAGE_SIZE, IMAGE_SIZE);
	pixels = gdk_pixbuf_get_pixels(pixbuf);
	surface_p = cairo_image_surface_create_for_data(pixels, CAIRO_FORMAT_ARGB32,
		IMAGE_SIZE, IMAGE_SIZE, gdk_pixbuf_get_rowstride(pixbuf));
	cairo_p = cairo_create(surface_p);
	cairo_set_source_rgb(cairo_p, 1, 1, 1);
	cairo_rectangle(cairo_p, 0, 0, IMAGE_SIZE, IMAGE_SIZE);
	cairo_fill(cairo_p);
	cairo_set_source_rgb(cairo_p, 0, 0, 0);
	{
		double root3 = sqrt(3.0);
		cairo_move_to(cairo_p, IMAGE_SIZE * 0.5, 0);
		cairo_line_to(cairo_p, IMAGE_SIZE * 0.5 - root3 * IMAGE_SIZE * 0.25, IMAGE_SIZE * 0.25);
		cairo_line_to(cairo_p, IMAGE_SIZE * 0.5 - root3 * IMAGE_SIZE * 0.25, (IMAGE_SIZE * 0.25) * 3.0);
		cairo_line_to(cairo_p, IMAGE_SIZE * 0.5, IMAGE_SIZE);
		cairo_line_to(cairo_p, IMAGE_SIZE * 0.5 + root3 * IMAGE_SIZE * 0.25, (IMAGE_SIZE * 0.25) * 3.0);
		cairo_line_to(cairo_p, IMAGE_SIZE * 0.5 + root3 * IMAGE_SIZE * 0.25, IMAGE_SIZE * 0.25);
		cairo_close_path(cairo_p);
		cairo_fill(cairo_p);
	}
	cairo_destroy(cairo_p);
	cairo_surface_destroy(surface_p);
	image = gtk_image_new_from_pixbuf(pixbuf);
	control = ToggleRadioButtonNew(&list,
		(void (*)(GtkWidget*, void*))OnClickedPatternButton, (void*)bokeh);
	g_object_set_data(G_OBJECT(control), "shape_type", GINT_TO_POINTER(BOKEH_SHAPE_TYPE_HEXAGON));
	gtk_container_add(GTK_CONTAINER(control), image);
	gtk_box_pack_start(GTK_BOX(layout), control, FALSE, FALSE, 0);

	// 䊐`
	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, IMAGE_SIZE, IMAGE_SIZE);
	pixels = gdk_pixbuf_get_pixels(pixbuf);
	surface_p = cairo_image_surface_create_for_data(pixels, CAIRO_FORMAT_ARGB32,
		IMAGE_SIZE, IMAGE_SIZE, gdk_pixbuf_get_rowstride(pixbuf));
	cairo_p = cairo_create(surface_p);
	cairo_set_source_rgb(cairo_p, 1, 1, 1);
	cairo_rectangle(cairo_p, 0, 0, IMAGE_SIZE, IMAGE_SIZE);
	cairo_fill(cairo_p);
	cairo_set_source_rgb(cairo_p, 0, 0, 0);
	{
		double points[5][2];
		GetPolygonPoints(points, 5, IMAGE_SIZE * 0.5, IMAGE_SIZE * 0.5, IMAGE_SIZE * 0.5);
		cairo_move_to(cairo_p, points[0][0], points[0][1]);
		for(i=1; i<5; i++)
		{
			cairo_line_to(cairo_p, points[(i*2)%5][0], points[(i*2)%5][1]);
		}
		cairo_fill(cairo_p);
	}
	cairo_destroy(cairo_p);
	cairo_surface_destroy(surface_p);
	image = gtk_image_new_from_pixbuf(pixbuf);
	control = ToggleRadioButtonNew(&list,
		(void (*)(GtkWidget*, void*))OnClickedPatternButton, (void*)bokeh);
	g_object_set_data(G_OBJECT(control), "shape_type", GINT_TO_POINTER(BOKEH_SHAPE_TYPE_STAR));
	gtk_container_add(GTK_CONTAINER(control), image);
	gtk_box_pack_start(GTK_BOX(layout), control, FALSE, FALSE, 0);

	return vbox;
#undef IMAGE_SIZE
}

#ifdef __cplusplus
}
#endif
