#include <stdlib.h>
#include <string.h>
#include "effect.h"
#include "bokeh.h"
#include "unsharp_mask.h"
#include "../../plug_in.h"
#include "../../utils.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef enum _eEFFECT_COLUMN_TYPE
{
	EFFECT_COLUMN_TYPE_NAME,
	EFFECT_COLUMN_TYPE_INDEX,
	NUM_EFFECT_COLUMN_TYPE
} eEFFECT_COLUMN_TYPE;

void InitializeEffectFilter(EFFECT* effect, LAYER* target)
{
	(void)memset(effect, 0, sizeof(*effect));
	effect->target = target;
}

static void SetEffectFilterData(
	EFFECT* effect,
	eEFFECT_TYPE type
)
{
	switch(type)
	{
	case EFFECT_TYPE_BOKEH:
		effect->effect_data = malloc(sizeof(BOKEH_DATA));
		InitializeBokeh((BOKEH_DATA*)effect->effect_data, effect->target);
		break;
	case EFFECT_TYPE_UNSHARP_MASK:
		effect->effect_data = malloc(sizeof(UNSHARP_MASK_DATA));
		InitializeUnsharpMask((UNSHARP_MASK_DATA*)effect->effect_data);
		break;
	}
	effect->type = type;
}

static int Progress(float rate, EFFECT* effect)
{
	if(rate >= effect->next)
	{
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(effect->target->window->app->progress), rate);
		UpdateWidget(effect->target->window->app->progress);
		effect->next += 0.02f;
	}

	return effect->canceled;
}

int ApplyEffectFilter(EFFECT* effect)
{
	effect->canceled = FALSE;
	effect->next = 0.0f;
	if(effect->effect_data != NULL)
	{
		switch(effect->type)
		{
		case EFFECT_TYPE_BOKEH:
			return Bokeh((BOKEH_DATA*)effect->effect_data, effect->target,
				(int (*)(float, void*))Progress, (void*)effect);
		case EFFECT_TYPE_UNSHARP_MASK:
			return UnsharpMask((UNSHARP_MASK_DATA*)effect->effect_data, effect->target,
				(int (*)(float, void*))Progress, (void*)effect);
		}
	}

	// vOXo[Zbg
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(effect->target->window->app->progress), 0);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(effect->target->window->app->progress), "");

	return 0;
}

static void ReleaseEffectFilterData(EFFECT* effect)
{
	if(effect->effect_data != NULL)
	{
		switch(effect->type)
		{
		case EFFECT_TYPE_BOKEH:
			ReleaseBokeh((BOKEH_DATA*)effect->effect_data);
			break;
		case EFFECT_TYPE_UNSHARP_MASK:
			break;
		}
		free(effect->effect_data);
		effect->effect_data = NULL;
	}
}

static GtkWidget* EffectDetailSettingWidgetNew(EFFECT* effect)
{
	switch(effect->type)
	{
	case EFFECT_TYPE_BOKEH:
		return BokehFilterSettingWidgetNew((BOKEH_DATA*)effect->effect_data, effect->target);
	case EFFECT_TYPE_UNSHARP_MASK:
		return UnsharpMaskSettingWidgetNew(
			(UNSHARP_MASK_DATA*)effect->effect_data, effect->target->window->app);
	}

	return NULL;
}

static void EffectItemChanged(GObject* selection, EFFECT* effect)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	eEFFECT_TYPE type;

	if(gtk_tree_selection_get_selected(GTK_TREE_SELECTION(selection), &model, &iter) != FALSE)
	{
		gtk_tree_model_get(model, &iter, 1, &type, -1);

		if(effect->setting != NULL)
		{
			gtk_widget_destroy(effect->setting);
			effect->setting = NULL;
		}

		ReleaseEffectFilterData(effect);
		SetEffectFilterData(effect, type);
		effect->setting = EffectDetailSettingWidgetNew(effect);

		gtk_box_pack_start(GTK_BOX(effect->box), effect->setting, TRUE, TRUE, 3);
		gtk_widget_show_all(effect->box);

		if(effect->setting != NULL)
		{
			gtk_box_pack_start(GTK_BOX(effect->box), effect->setting, FALSE, FALSE, 0);
			gtk_widget_show_all(effect->setting);
		}
	}
}

static void SetEffectTreeView(
	GtkWidget* tree,
	EFFECT* effect,
	APPLICATION* app
)
{
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkListStore *store;
	GtkTreeIter iter;
	GtkTreeSelection *selection;

	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree), FALSE);

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_add_attribute(column, renderer, "text", 0);
	(void)gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

	store = gtk_list_store_new(NUM_EFFECT_COLUMN_TYPE, G_TYPE_STRING, G_TYPE_INT);

	gtk_tree_view_set_model(GTK_TREE_VIEW(tree), GTK_TREE_MODEL(store));
	g_object_unref(store);

	store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(tree)));

	gtk_list_store_append(store, &iter);
	gtk_list_store_set(store, &iter, EFFECT_COLUMN_TYPE_NAME, "Bokeh",
		EFFECT_COLUMN_TYPE_INDEX, EFFECT_TYPE_BOKEH, -1);

	gtk_list_store_append(store, &iter);
	gtk_list_store_set(store, &iter, EFFECT_COLUMN_TYPE_NAME, "Unsharp Mask",
		EFFECT_COLUMN_TYPE_INDEX, EFFECT_TYPE_UNSHARP_MASK, -1);

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
	(void)g_signal_connect(G_OBJECT(selection), "changed",
		G_CALLBACK(EffectItemChanged), effect);
}

GtkWidget* EffectSettingWidgetNew(EFFECT* effect)
{
	GtkWidget *tree_view;

	effect->box = gtk_hbox_new(FALSE, 0);
	tree_view = gtk_tree_view_new();
	SetEffectTreeView(tree_view, effect, effect->target->window->app);

	gtk_box_pack_start(GTK_BOX(effect->box), tree_view, FALSE, FALSE, 0);

	return effect->box;
}

#ifdef __cplusplus
}
#endif
