#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "fft.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifndef M_PI
# define M_PI 3.14159265358979323846
#endif

void InitializeFFT(FFT_DATA* fft, int log_n)
{
	double angle;
	int n = 1;
	int i;

	(void)memset(fft, 0, sizeof(*fft));
	fft->weight1 = (float*)calloc(sizeof(*fft->weight1), log_n);
	fft->weight2 = (float*)calloc(sizeof(*fft->weight2), log_n);
	fft->weight3 = (float*)calloc(sizeof(*fft->weight3), log_n);

	for(i=0; i<log_n; i++)
	{
		n <<= 1;
		angle = - 2.0 * M_PI / n;
		fft->weight1[i] = (float)sin(0.5 * angle);
		fft->weight2[i] = -2.0f * fft->weight1[i] * fft->weight1[i];
		fft->weight3[i] = (float)sin(angle);
	}
}

void ReleaseFFT(FFT_DATA* fft)
{
	free(fft->weight1);
	free(fft->weight2);
	free(fft->weight3);
}

static void Scramble(int n, float* real, float* image)
{
	float t;
	int m;
	int i, j = 0;

	for(i=0; i<n; i++)
	{
		if(i > j)
		{
			t = real[j];
			real[j] = real[i];
			real[i] = t;
			t = image[j];
			image[j] = image[i];
			image[i] = t;
		}
		m = n >> 1;
		while(j >= m && m >= 2)
		{
			j -= m;
			m >>= 1;
		}
		j += m;
	}
}

static void Butterflies(
	FFT_DATA* fft,
	int n,
	int log_n,
	int direction,
	float* real,
	float* image
)
{
	int m = 1;
	int i, j, k;

	for(i=0; i<log_n; i++)
	{
		float weight_real, weight_image;
		float weight_point_real, weight_point_image;
		float temp_real, temp_image;
		float weight;
		int half_n = m;
		int offset;
		m <<= 1;
		weight = direction * fft->weight1[i];
		weight_point_real = fft->weight2[i];
		weight_point_image = direction * fft->weight3[i];
		weight_real = 1.0f;
		weight_image = 0.0f;
		for(offset=0; offset<half_n; offset++)
		{
			for(j=offset; j<n; j += m)
			{
				float real_value, image_value;
				k = j + half_n;
				real_value = real[k];
				image_value = image[k];
				temp_real = (weight_real * real_value) - (weight_image * image_value);
				temp_image = (weight_image * real_value) + (weight_real * image_value);
				real[k] = real[j] - temp_real;
				real[j] += temp_real;
				image[k] = image[j] - temp_image;
				image[j] += temp_image;
			}
			weight = weight_real;
			weight_real = weight * weight_point_real - weight_image * weight_point_image + weight_real;
			weight_image = weight_image * weight_point_real + weight * weight_point_image + weight_image;
		}
	}

	if(direction < 0)
	{
		float nr = 1.0f / n;
		for(i=0; i<n; i++)
		{
			real[i] *= nr;
			image[i] *= nr;
		}
	}
}

static int Log2(int n)
{
	int m = 1;
	int log2_n = 0;

	while(m < n)
	{
		m *= 2;
		log2_n++;
	}

	return m == n ? log2_n : -1;
}

void FastFourierTransform1D(
	FFT_DATA* fft,
	float* real,
	float* image,
	int log_n,
	int n,
	int direction
)
{
	Scramble(n, real, image);
	Butterflies(fft, n, log_n, direction, real, image);
}

void FastFourierTransform2D(
	FFT_DATA* fft,
	float* real,
	float* image,
	int columns,
	int rows,
	int direction
)
{
	int log2_columns = Log2(columns);
	int log2_rows = Log2(rows);
	int n = (rows > columns) ? rows : columns;
	float *real_temp = (float*)malloc(sizeof(*real_temp) * n);
	float *image_temp = (float*)malloc(sizeof(*image_temp) * n);
	int x, y;

	for(y=0; y<rows; y++)
	{
		int offset = y * columns;
		(void)memcpy(real_temp, &real[offset], sizeof(*real)*columns);
		(void)memcpy(image_temp, &image[offset], sizeof(*image)*columns);
		FastFourierTransform1D(fft, real_temp, image_temp, log2_columns, columns, direction);
		(void)memcpy(&real[offset], real_temp, sizeof(*real)*columns);
		(void)memcpy(&image[offset], image_temp, sizeof(*image)*columns);
	}

	for(x=0; x<columns; x++)
	{
		int index = x;
		for(y=0; y<rows; y++)
		{
			real_temp[y] = real[index];
			image_temp[y] = image[index];
			index += columns;
		}
		FastFourierTransform1D(fft, real_temp, image_temp, log2_rows, rows, direction);
		index = x;
		for(y=0; y<rows; y++)
		{
			real[index] = real_temp[y];
			image[index] = image_temp[y];
			index += columns;
		}
	}

	free(real_temp);
	free(image_temp);
}

#ifdef __cplusplus
}
#endif
