/* graph.cpp */
/* Vladimir Kolmogorov (vnk@cs.cornell.edu), 2001. */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "graph.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifndef INLINE
# ifdef _MSC_VER
#  define INLINE __inline
# else
#  define INLINE inline
# endif
#endif

#define MAX_CAPACITY 16383
#define REMPLI 1

#ifndef MAXIMUM
# define MAXIMUM(A, B) (((A) > (B)) ? (A) : (B))
#endif

#ifndef MINIMUM
# define MINIMUM(A, B) (((A) < (B)) ? (A) : (B))
#endif

#define NODE_BLOCK_SIZE 512
#define ARC_BLOCK_SIZE 1024
#define NODEPTR_BLOCK_SIZE 128

#define TERMINAL ((GRAPH_ARC*)1)
#define ORPHAN ((GRAPH_ARC*)2)

#define INFINITE_DISTANCE 1000000000

void InitializeGraph(GRAPH* graph, void (*error_function)(char*))
{
	(void)memset(graph, 0, sizeof(*graph));
	graph->error_function = error_function;
	InitializeBlock(&graph->node_block, NODE_BLOCK_SIZE, sizeof(GRAPH_NODE), error_function);
	InitializeBlock(&graph->arc_block, ARC_BLOCK_SIZE, sizeof(GRAPH_ARC), error_function);
}

void ReleaseGraph(GRAPH* graph)
{
	ReleaseBlock(&graph->node_block);
	ReleaseBlock(&graph->arc_block);
}

void* GraphAddNode(GRAPH* graph)
{
	GRAPH_NODE *node = (GRAPH_NODE*)BlockAppend(&graph->node_block, 1);

	node->first = NULL;
	node->tr_cap = 0;

	return (void*)node;
}

void GraphAddEdge(
	GRAPH* graph,
	void* from,
	void* to,
	short cap,
	short rev_cap
)
{
	GRAPH_ARC *arc, *arc_reverse;

	arc = (GRAPH_ARC*)BlockAppend(&graph->arc_block, 2);
	arc_reverse = arc + 1;

	arc->brother = arc_reverse;
	arc_reverse->brother = arc;
	arc->next = ((GRAPH_NODE*)from)->first;
	((GRAPH_NODE*)from)->first = arc;
	arc_reverse->next = ((GRAPH_NODE*)to)->first;
	((GRAPH_NODE*)to)->first = arc_reverse;
	arc->head = (GRAPH_NODE*)to;
	arc_reverse->head = (GRAPH_NODE*)from;
	arc->r_cap = cap;
	arc_reverse->r_cap = rev_cap;
}

void GraphSetWeights(GRAPH* graph, void* node, short cap_source, short cap_sink)
{
	graph->flow += (cap_source < cap_sink) ? cap_source : cap_sink;
	((GRAPH_NODE*)node)->tr_cap = cap_source - cap_sink;
}

void GraphAddWeights(GRAPH* graph, void* node, short cap_source, short cap_sink)
{
	register short delta = ((GRAPH_NODE*)node)->tr_cap;
	if(delta > 0)
	{
		cap_source += delta;
	}
	else
	{
		cap_sink -= delta;
	}
	graph->flow += (cap_source < cap_sink) ? cap_source : cap_sink;
	((GRAPH_NODE*)node)->tr_cap = cap_source - cap_sink;
}

static INLINE void GraphSetActive(GRAPH* graph, GRAPH_NODE* node)
{
	if(node->next == NULL)
	{
		if(graph->queue_last[1] != NULL)
		{
			graph->queue_last[1]->next = node;
		}
		else
		{
			graph->queue_first[1] = node;
		}
		graph->queue_last[1] = node;
		node->next = node;
	}
}

static INLINE GRAPH_NODE* GraphNextActive(GRAPH* graph)
{
	GRAPH_NODE *node;

	while(1)
	{
		if((node=graph->queue_first[0]) == NULL)
		{
			graph->queue_first[0] = node = graph->queue_first[1];
			graph->queue_last[0]  = graph->queue_last[1];
			graph->queue_first[1] = NULL;
			graph->queue_last[1]  = NULL;
			if(node == NULL)
			{
				return NULL;
			}
		}

		if(node->next == node)
		{
			graph->queue_first[0] = graph->queue_last[0] = NULL;
		}
		else
		{
			graph->queue_first[0] = node->next;
		}
		node->next = NULL;

		if(node->parent)
		{
			return node;
		}
	}

	return NULL;
}

void InitializeGraphMaxflow(GRAPH* graph)
{
	GRAPH_NODE *node;

	graph->queue_first[0] = graph->queue_last[0] = NULL;
	graph->queue_first[1] = graph->queue_last[1] = NULL;
	graph->orphan_first = NULL;

	for (node=(GRAPH_NODE*)BlockScanFirst(&graph->node_block); node != NULL;
		node=(GRAPH_NODE*)BlockScanNext(&graph->node_block))
	{
		node->next = NULL;
		node->time_stamp = 0;
		if(node->tr_cap > 0)
		{
			node->is_sink = 0;
			node->parent = TERMINAL;
			GraphSetActive(graph, node);
			node->time_stamp = 0;
			node->distance = 1;
		}
		else if(node->tr_cap < 0)
		{
			node->is_sink = 1;
			node->parent = TERMINAL;
			GraphSetActive(graph, node);
			node->time_stamp = 0;
			node->distance = 1;
		}
		else
		{
			node->parent = NULL;
		}
	}
	graph->time_value = 0;
}

void GraphAugment(GRAPH* graph, GRAPH_ARC* middle_arc)
{
	GRAPH_NODE *node;
	GRAPH_ARC *arc;
	short bottleneck;
	GRAPH_NODE_POINTER *node_pointer;

	bottleneck = middle_arc->r_cap;
	for(node=middle_arc->brother->head; ; node=arc->head)
	{
		arc = node->parent;
		if(arc == TERMINAL)
		{
			break;
		}
		if(bottleneck > arc->brother->r_cap)
		{
			bottleneck = arc->brother->r_cap;
		}
	}
	if(bottleneck > node->tr_cap)
	{
		bottleneck = node->tr_cap;
	}

	for(node=middle_arc->head; ; node=arc->head)
	{
		arc = node->parent;
		if(arc == TERMINAL)
		{
			break;
		}
		if(bottleneck > arc->r_cap)
		{
			bottleneck = arc->r_cap;
		}
	}
	if(bottleneck > - node->tr_cap)
	{
		bottleneck = - node->tr_cap;
	}

	middle_arc->brother->r_cap += bottleneck;
	middle_arc->r_cap -= bottleneck;
	for (node=middle_arc->brother->head; ; node=arc->head)
	{
		arc = node->parent;
		if(arc == TERMINAL)
		{
			break;
		}
		arc->r_cap += bottleneck;
		arc->brother->r_cap -= bottleneck;
		if(arc->brother->r_cap == 0)
		{
			node->parent = ORPHAN;
			node_pointer = (GRAPH_NODE_POINTER*)DeleteBlockAppend(&graph->node_pointer_block);
			node_pointer->pointer = node;
			node_pointer->next = graph->orphan_first;
			graph->orphan_first = node_pointer;
		}
	}
	node->tr_cap -= bottleneck;
	if(node->tr_cap == 0)
	{
		node->parent = ORPHAN;
		node_pointer = (GRAPH_NODE_POINTER*)DeleteBlockAppend(&graph->node_pointer_block);
		node_pointer->pointer = node;
		node_pointer->next = graph->orphan_first;
		graph->orphan_first = node_pointer;
	}

	for(node=middle_arc->head; ; node=arc->head)
	{
		arc = node->parent;
		if(arc == TERMINAL)
		{
			break;
		}
		arc->brother->r_cap += bottleneck;
		arc->r_cap -= bottleneck;
		if(arc->r_cap == 0)
		{
			node->parent = ORPHAN;
			node_pointer = (GRAPH_NODE_POINTER*)DeleteBlockAppend(&graph->node_pointer_block);
			node_pointer->pointer = node;
			node_pointer->next = graph->orphan_first;
			graph->orphan_first = node_pointer;
		}
	}
	node->tr_cap += bottleneck;
	if(node->tr_cap == 0)
	{
		node->parent = ORPHAN;
		node_pointer = (GRAPH_NODE_POINTER*)DeleteBlockAppend(&graph->node_pointer_block);
		node_pointer->pointer = node;
		node_pointer->next = graph->orphan_first;
		graph->orphan_first = node_pointer;
	}

	graph->flow += bottleneck;
}

void GraphProcessSourceOrphan(GRAPH* graph, GRAPH_NODE* i)
{
	GRAPH_NODE *j;
	GRAPH_ARC *arc0, *arc0_minimum = NULL, *arc;
	GRAPH_NODE_POINTER *node_pointer;
	int distance, distance_minimum = INFINITE_DISTANCE;

	for(arc0=i->first; arc0; arc0=arc0->next)
	if(arc0->brother->r_cap)
	{
		j = arc0 -> head;
		if(j->is_sink == 0 && (arc=j->parent) != NULL)
		{
			distance = 0;
			while (1)
			{
				if(j->time_stamp == graph->time_value)
				{
					distance += j->distance;
					break;
				}
				arc = j->parent;
				distance++;
				if(arc == TERMINAL)
				{
					j->time_stamp = graph->time_value;
					j->distance = 1;
					break;
				}
				if(arc == ORPHAN)
				{
					distance = INFINITE_DISTANCE;
					break;
				}
				j = arc->head;
			}
			if(distance < INFINITE_DISTANCE)
			{
				if(distance < distance_minimum)
				{
					arc0_minimum = arc0;
					distance_minimum = distance;
				}

				for(j=arc0->head; j->time_stamp != graph->time_value; j=j->parent->head)
				{
					j->time_stamp = graph->time_value;
					j->distance = distance--;
				}
			}
		}
	}

	if((i->parent = arc0_minimum) != NULL)
	{
		i->time_stamp = graph->time_value;
		i->distance = distance_minimum + 1;
	}
	else
	{
		i->time_stamp = 0;

		for(arc0=i->first; arc0 != NULL; arc0=arc0->next)
		{
			j = arc0->head;
			if(j->is_sink == 0 && (arc = j->parent) != NULL)
			{
				if(arc0->brother->r_cap)
				{
					GraphSetActive(graph, j);
				}
				if(arc != TERMINAL && arc != ORPHAN && arc->head == i)
				{
					j->parent = ORPHAN;
					node_pointer = (GRAPH_NODE_POINTER*)DeleteBlockAppend(&graph->node_pointer_block);
					node_pointer->pointer = j;
					if(graph->orphan_last != NULL)
					{
						graph->orphan_last->next = node_pointer;
					}
					else
					{
						graph->orphan_first = node_pointer;
					}
					graph->orphan_last = node_pointer;
					node_pointer->next = NULL;
				}
			}
		}
	}
}

void GraphProcessSinkOrphan(GRAPH* graph, GRAPH_NODE* i)
{
	GRAPH_NODE *j;
	GRAPH_ARC *arc0, *arc0_minimum = NULL, *arc;
	GRAPH_NODE_POINTER *node_pointer;
	int distance, distance_minimum = INFINITE_DISTANCE;

	for(arc0=i->first; arc0; arc0=arc0->next)
	if(arc0->r_cap != 0)
	{
		j = arc0->head;
		if(j->is_sink != 0 && (arc = j->parent) != NULL)
		{
			distance = 0;
			while(1)
			{
				if(j->time_stamp == graph->time_value)
				{
					distance += j->distance;
					break;
				}
				arc = j->parent;
				distance++;
				if(arc == TERMINAL)
				{
					j->time_stamp = graph->time_value;
					j->distance = 1;
					break;
				}
				if(arc == ORPHAN)
				{
					distance = INFINITE_DISTANCE;
					break;
				}
				j = arc->head;
			}
			if(distance < INFINITE_DISTANCE)
			{
				if(distance < distance_minimum)
				{
					arc0_minimum = arc0;
					distance_minimum = distance;
				}

				for(j=arc0->head; j->time_stamp != graph->time_value; j=j->parent->head)
				{
					j->time_stamp = graph->time_value;
					j->distance = distance--;
				}
			}
		}
	}

	if((i->parent = arc0_minimum) != NULL)
	{
		i->time_stamp = graph->time_value;
		i->distance = distance_minimum + 1;
	}
	else
	{
		i->time_stamp = 0;

		for(arc0=i->first; arc0 != NULL; arc0=arc0->next)
		{
			j = arc0->head;
			if(j->is_sink != 0 && (arc = j->parent) != NULL)
			{
				if(arc0->r_cap != 0)
				{
					GraphSetActive(graph, j);
				}
				if(arc != TERMINAL && arc != ORPHAN && arc->head == i)
				{
					j->parent = ORPHAN;
					node_pointer = (GRAPH_NODE_POINTER*)DeleteBlockAppend(&graph->node_pointer_block);
					node_pointer->pointer = j;
					if(graph->orphan_last != NULL)
					{
						graph->orphan_last->next = node_pointer;
					}
					else
					{
						graph->orphan_first = node_pointer;
					}
					graph->orphan_last = node_pointer;
					node_pointer->next = NULL;
				}
			}
		}
	}
}

int GraphMaxFlow(GRAPH* graph)
{
	GRAPH_NODE *i, *j, *current_node = NULL;
	GRAPH_ARC *arc;
	GRAPH_NODE_POINTER *node_pointer, *node_pointer_next;

	InitializeGraphMaxflow(graph);
	InitializeDeleteBlock(&graph->node_pointer_block, NODEPTR_BLOCK_SIZE,
		sizeof(GRAPH_NODE_POINTER), graph->error_function);

	while(1)
	{
		if (i=current_node)
		{
			i->next = NULL;
			if(i->parent == NULL)
			{
				i = NULL;
			}
		}
		if(i == NULL)
		{
			if((i = GraphNextActive(graph)) == NULL)
			{
				break;
			}
		}

		if(i->is_sink == 0)
		{
			for(arc=i->first; arc != NULL; arc=arc->next)
			if(arc->r_cap != 0)
			{
				j = arc->head;
				if(j->parent == NULL)
				{
					j->is_sink = 0;
					j->parent = arc->brother;
					j->time_stamp = i->time_stamp;
					j->distance = i->distance + 1;
					GraphSetActive(graph, j);
				}
				else if(j->is_sink != 0)
				{
					break;
				}
				else if(j->time_stamp <= i->time_stamp &&
					j->distance > i->distance)
				{
					j->parent = arc->brother;
					j->time_stamp = i->time_stamp;
					j->distance = i->distance + 1;
				}
			}
		}
		else
		{
			for (arc=i->first; arc != NULL; arc=arc->next)
			if(arc->brother->r_cap)
			{
				j = arc->head;
				if(j->parent == NULL)
				{
					j->is_sink = 1;
					j->parent = arc->brother;
					j->time_stamp = i->time_stamp;
					j->distance = i->distance + 1;
					GraphSetActive(graph, j);
				}
				else if(j->is_sink == 0)
				{
					arc = arc->brother;
					break;
				}
				else if (j->time_stamp <= i->time_stamp &&
					j->distance > i->distance)
				{
					j->parent = arc->brother;
					j->time_stamp = i->time_stamp;
					j->distance = i->distance + 1;
				}
			}
		}

		graph->time_value++;

		if(arc != NULL)
		{
			i->next = i;
			current_node = i;

			GraphAugment(graph, arc);

			while((node_pointer = graph->orphan_first) != NULL)
			{
				node_pointer_next = node_pointer->next;
				node_pointer->next = NULL;

				while((node_pointer = graph->orphan_first) != NULL)
				{
					graph->orphan_first = node_pointer->next;
					i = node_pointer->pointer;
					DeleteBlockRemove(&graph->node_pointer_block, (void*)node_pointer);
					if(graph->orphan_first == NULL)
					{
						graph->orphan_last = NULL;
					}
					if(i->is_sink != 0)
					{
						GraphProcessSinkOrphan(graph, i);
					}
					else
					{
						GraphProcessSourceOrphan(graph, i);
					}
				}

				graph->orphan_first = node_pointer_next;
			}
		}
		else
		{
			current_node = NULL;
		}
	}

	ReleaseDeleteBlock(&graph->node_pointer_block);

	return graph->flow;
}

eGRAPH_TERMINATE_TYPE GraphWhatSegment(void* node)
{
	if(((GRAPH_NODE*)node)->parent && ((GRAPH_NODE*)node)->is_sink == 0)
	{
		return GRAPH_TERMINATE_TYPE_SOURCE;
	}
	return GRAPH_TERMINATE_TYPE_SINK;
}

static INLINE short Cost(unsigned char* pixel1, unsigned char* pixel2, int channels)
{
	int difference, result = 0;
	int c;
	for(c = 0; c < channels; c++)
	{
		difference = pixel1[c] - pixel2[c];
		result += difference * difference;
	}
	return (result/24);
}

static INLINE short Gradient(unsigned char* pixel1, unsigned char* pixel2, int channels)
{
	int difference, result = 0;
	int c;
	for(c = 0; c < channels; c++)
	{
		difference = pixel1[c] - pixel2[c];
		result += difference * difference;
	}
	return ((short)sqrt((double)result));
}

static INLINE short EdgeWeight(
	int channels,
	unsigned char* im1_pix1,
	unsigned char* im2_pix1,
	unsigned char* im1_pix2,
	unsigned char* im2_pix2
)
{
	return ((Cost(im1_pix1,im2_pix1,channels) + (Cost(im1_pix2,im2_pix2,channels)))
		/ (Gradient(im1_pix1,im1_pix2,channels) + Gradient(im2_pix1,im2_pix2,channels) + 1));
}

static INLINE void PastePatchPixel2Image(
	int width_out,
	int height_out,
	int width_in,
	int height_in,
	int x,
	int y,
	int x_patch,
	int y_patch,
	int channels,
	unsigned char* image,
	unsigned char* patch
)
{
	int i;
	for(i = 0; i < channels; i++)
	{
		image[(y * width_out + x) * channels + i] = patch[(y_patch * width_in + x_patch) * channels + i];
	}
}

int CompterRemplis(unsigned char** rempli, int width_out, int height_out)
{
	int x, y;
	int somme = 0;

	for (x = 0; x < width_out; x++)
	{
		for (y = 0; y < height_out; y++) {
		if(rempli[x][y])
		{
			somme++;
		}
		}
	}
	return somme;
}

int* PixelRemplir(unsigned char** rempli, int width_out, int height_out, int* result_at)
{
	int x, y;

	for(y = 0; y < height_out; y++)
	{
		for(x = 0; x < width_out; x++)
		{
			if(rempli[x][y] == 0)
			{
				result_at[0] = x;
				result_at[1] = y;
				return result_at;
			}
		}
	}

	return NULL;
}

int Modulo(int x, int m)
{
	if(x >= m)
	{
		return x - m;
	}
	else
	{
		if(x >= 0)
		{
			return x;
		}
	}
	return x + m;
}

unsigned char** RempliNew(int x, int y)
{
	unsigned char** tab;
	unsigned char* pointer;
	int i;

	tab = (unsigned char**)malloc(x * sizeof (unsigned char*));
	tab[0] = (unsigned char*)malloc(y * x * sizeof(unsigned char));
	(void)memset(tab[0], 0, y * x * sizeof(unsigned char));

	pointer = tab[0] + y;
	for(i = 1; i < x; i++, pointer += y)
	{
		tab[i] = pointer;
	}

	return tab;
}

void DecoupeGraph(
	int* patch_position,
	int width_out,
	int height_out,
	int width_in,
	int height_in,
	int channels,
	unsigned char** rempli,
	unsigned char* image,
	unsigned char* patch,
	unsigned char* coupe_h_here,
	unsigned char* coupe_h_west,
	unsigned char* coupe_v_here,
	unsigned char* coupe_v_north,
	int* compter,
	int make_tileable,
	int invert,
	void (*error_function)(char*)
)
{
	int x_patch, y_patch, x, y;
	int real_x, real_y;
	int x_inf, y_inf, x_sup, y_sup;
	GRAPH graph;
	void **node_of_pixel;
	short poids;
	void *first_node = NULL;
	void *node_sommet_courant; 
	unsigned char r;

	InitializeGraph(&graph, error_function);
	node_of_pixel = (void**)calloc(sizeof(void*), width_in * height_in);

	if(make_tileable)
	{
		x_inf = patch_position[0];
		y_inf = patch_position[1];
		x_sup = patch_position[0] + width_in;
		y_sup = patch_position[1] + height_in;
	}
	else
	{
		x_inf = MAXIMUM(0, patch_position[0]);
		y_inf = MAXIMUM(0, patch_position[1]);
		x_sup = MINIMUM(width_out,  patch_position[0] + width_in);
		y_sup = MINIMUM(height_out, patch_position[1] + height_in);
	}

	for(real_x = x_inf; real_x < x_sup; real_x++)
	{
		x_patch = real_x - patch_position[0];
		x = Modulo(real_x, width_out);
		for(real_y = y_inf; real_y < y_sup; real_y++)
		{
			y_patch = real_y - patch_position[1];
			y = Modulo(real_y, height_out);

			if(rempli[x][y])
			{
				node_of_pixel[x_patch * height_in + y_patch] = GraphAddNode(&graph);
				if(first_node == NULL)
				{
					first_node = node_of_pixel[x_patch * height_in + y_patch];
				}
			}
		}
	}

	for(real_x = x_inf; real_x < x_sup; real_x++)
	{
		x_patch = real_x - patch_position[0];
		x = Modulo(real_x, width_out);

		for(real_y = y_inf; real_y < y_sup; real_y++)
		{
			y_patch = real_y - patch_position[1];
			y = Modulo(real_y, height_out);

			if(rempli[x][y] == 0)
			{
				continue;
			}
			else
			{
				node_sommet_courant = node_of_pixel[x_patch * height_in + y_patch];

				if((!make_tileable && y_patch != 0 && y != 0 && rempli[x][y - 1])
					|| (make_tileable && y_patch != 0 && rempli[x][Modulo (y - 1, height_out)]))
				{
					poids = EdgeWeight (channels,
						image + ((y * width_out + x) * channels),
						patch + ((y_patch * width_in + x_patch) * channels),
						image + (((Modulo(y - 1, height_out)) * width_out + x) * channels),
						patch + (((y_patch - 1) * width_in + x_patch) * channels)
					);
					GraphAddEdge(&graph, node_sommet_courant,
						node_of_pixel[x_patch * height_in + y_patch - 1], poids, poids);
				}

				if((!make_tileable && x_patch != 0 && x != 0 && rempli[x - 1][y])
					|| (make_tileable && x_patch != 0 && rempli[Modulo (x - 1, width_out)][y]))
				{
					poids = EdgeWeight(channels,
						image + ((y * width_out + x) * channels),
						patch + ((y_patch * width_in + x_patch) * channels),
						image + ((y * width_out + (Modulo(x, width_out) - 1)) * channels),
						patch + ((y_patch * width_in + (x_patch - 1)) * channels)
					);
					GraphAddEdge(&graph, node_sommet_courant,
						node_of_pixel[(x_patch - 1) * height_in + y_patch], poids, poids);
				}

				if((make_tileable && (x_patch == 0 || y_patch == 0 || x_patch == width_in - 1 || y_patch == height_in - 1))
					|| (!make_tileable && (x_patch == 0 || y_patch == 0 || x_patch == width_in - 1 || y_patch == height_in - 1)
						&& x != 0 && y != 0 && x != width_out - 1 && y != height_out - 1))
				{
					GraphAddWeights(&graph, node_sommet_courant, MAX_CAPACITY, 0);
				}

				if(((!make_tileable)
					&& ((y_patch != 0 && y != 0 && !rempli[x][y - 1])
						|| (y_patch != height_in - 1 && y != height_out - 1 && rempli[x][y + 1] == 0)
						|| (x_patch != width_in - 1  && x != width_out - 1  && rempli[x + 1][y] == 0)
						|| (x_patch != 0 && x != 0 && !rempli[x - 1][y] == 0)))
					|| ((make_tileable)
					&& ((y_patch != 0 && rempli[x][Modulo(y - 1, height_out)] == 0)
						|| (y_patch != height_in - 1 && rempli[x][Modulo(y + 1, height_out)] == 0)
						|| (x_patch != width_in - 1  && rempli[Modulo(x + 1, width_out)][y] == 0)
						|| (x_patch != 0 && rempli[Modulo(x - 1, width_out)][y] == 0))))
				{
					GraphAddWeights(&graph, node_sommet_courant, 0, MAX_CAPACITY);
				}
			}
		}
	}

	if(!make_tileable)
	{
		GraphAddWeights(&graph, first_node, MAX_CAPACITY, 0);
	}

	(void)GraphMaxFlow(&graph);

	for(real_x = x_inf; real_x < x_sup; real_x++)
	{
		x_patch = real_x - patch_position[0];
		x = Modulo(real_x, width_out);
		for(real_y = y_inf; real_y < y_sup; real_y++)
		{
			y_patch = real_y - patch_position[1];
			y = Modulo(real_y, height_out);
			r = rempli[x][y];
			if(r != 0)
			{
				if(GraphWhatSegment(node_of_pixel[x_patch * height_in + y_patch]) == GRAPH_TERMINATE_TYPE_SINK)
				{
					PastePatchPixel2Image(width_out, height_out, width_in, height_in, x, y,
						x_patch, y_patch, channels, image, patch);
				}
			}
			else
			{ // (rempli[x_i][y_i] != 0)
				PastePatchPixel2Image(width_out, height_out, width_in, height_in, x, y,
					x_patch, y_patch, channels, image, patch);
				rempli[x][y] = REMPLI;
				*compter = *compter + 1;
			}
		}
	}

	ReleaseGraph(&graph);
	free(node_of_pixel);
}

static double Difference(
	int width_out,
	int height_out,
	int width_in,
	int height_in,
	unsigned char* image,
	unsigned char* patch,
	int position_x,
	int position_y,
	int x_minimum,
	int y_minimum,
	int x_maximum,
	int y_maximum,
	int channels,
	unsigned char** rempli
)
{
	int somme = 0, zone=0, x, y, i;
	unsigned char *image_ptr, *patch_ptr;
	int real_x, real_y;
	int modulo_value;

	for(y = y_minimum; y < y_maximum; y++)
	{
		real_y = Modulo(y, height_out);

		modulo_value = Modulo(x_minimum, width_out);
		image_ptr = image + (real_y * width_out
			+ modulo_value) * channels;
		patch_ptr = patch + ((real_y - position_y) * width_in
			+ modulo_value - position_x) * channels;

		for(x = x_minimum; x < x_maximum; x++)
		{
			real_x = Modulo(x, width_out);

			if(rempli[real_x][real_y] != 0)
			{
				for(i = 0 ; i < channels; i++)
				{
					somme += abs(*image_ptr++ - *patch_ptr++);
					zone ++;
				}
			}
			else
			{
				image_ptr += channels;
				patch_ptr += channels;
			}
		}
	}

	if(zone == 0)
	{
		return -1;
	}
	return (((double)somme) / ((double)zone));
}

void OffsetOptimal(
	int* result_at,
	unsigned char* image,
	unsigned char* patch,
	int width_in,
	int height_in,
	int width_out,
	int height_out,
	int x_patch_position_minimum,
	int y_patch_position_minimum,
	int x_patch_position_maximum,
	int y_patch_position_maximum,
	int channels,
	unsigned char** rempli,
	int tileable
)
{
	int x, y;
	double best_difference = HUGE, temp_difference;

	for(x = x_patch_position_minimum; x < x_patch_position_maximum; x++)
	{
		for (y = y_patch_position_minimum; y < y_patch_position_maximum; y++)
		{
			if(tileable)
			{
				temp_difference = Difference(
					width_out, height_out,
					width_in, height_in,
					image, patch,
					x, y,
					MAXIMUM(0, x), MAXIMUM(0, y), x + width_in, y + height_in,
					channels, rempli
				);
			}
			else
			{
				temp_difference = Difference(
					width_out, height_out,
					width_in, height_in,
					image, patch,
					x, y,
					MAXIMUM(0,x), MAXIMUM(0,y), MINIMUM(x + width_in, width_out), MINIMUM(y + height_in, height_out),
					channels, rempli
				);
			}
			if(temp_difference < best_difference)
			{
				best_difference = temp_difference;
				result_at[0] = x;	result_at[1] = y;
			}
		}
	}
}

#ifdef __cplusplus
}
#endif

